#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <pthread.h>
#include <sndfile.h>
#include <samplerate.h>
#include <libxml/parser.h>
#include <jack/types.h>

#include "specimen.h"
#include "patch.h"
#include "driver.h"

#define RINGMAX 1024 /* arbitrary, is there a better way? */

typedef enum event_s { EV_OFF=0, EV_PLAY, EV_STOP } event_t;

typedef struct sample_s {
     pthread_mutex_t mutex;             /* never gets initialized, somehow works... shh... */
     float *sp;                         /* samples pointer */

     char name[PATCH_MAX_SAMPLE_NAME];  /* name of sample file */
     int frames;                        /* number of frames (not samples) pointed to by *sp */
} sample_t;

/* type for array of instruments (called patches) */
typedef struct patch_s {
     sample_t sample;
     int display_index;                 /* order in which it is to be displayed */
     char name[PATCH_MAX_NAME];         /* string name */
     int channel;                       /* midi channel to listen on */
     int note;                          /* midi note to listen on */
     int range;                         /* whether to listen to a range of notes or not */
     int lower_note;                    /* lowest note in range */
     int upper_note;                    /* highest note in range */
     float volume;
     float pan;
     int active;                        /* whether this instance of patch_s is in use or not */
     play_mode_t play_mode;
     int cut;
     int cut_by;
     int sample_start;
     int sample_stop;
     int loop_start;
     int loop_stop;
} patch_t;

/* type for array of currently playing notes  */
typedef struct patch_playlist_s {
     patch_t *patch;
     int note;
     float velocity;
     int offset;
     int next_frame;
     double step;        /* number of frames provided per frames requested */
     double delta;       /* distance between next_frame and next_frame+1 */
     int active;
     int tailo;          /* number of frames until STOP fade starts (tail offset) */
     int tailc;          /* number of frames left in fade out on STOP events (tail count)*/
     float tailv;        /* current volume for fades on STOP events (tail volume) */
} patch_playlist_t;

/* type for ringbuffer of incoming events */
typedef struct patch_ring_s {
     patch_t *patch;
     tick_t tick;
     int event;
     float velocity;
     int note;
} patch_ring_t;

/* special structure for previewing samples */
typedef struct preview_s {
     sample_t sample;
     int active;
     int next_frame;
} preview_t;

static patch_t patch[PATCH_COUNT];
static preview_t preview;
static patch_playlist_t patch_playlist[PATCH_POLYPHONY];
static patch_ring_t patch_ring[RINGMAX];
static patch_ring_t *patch_ring_feeder = patch_ring;
static patch_ring_t *patch_ring_consumer = patch_ring;

/* guess what this does to my sphincter */
static float master_volume;

/* saves some typing */
const char *patch_strerror(int error)
{
     switch (error) {
     case PATCH_ID_INVALID:
	  return "patch id is invalid";
	  break;
     case PATCH_EMPTY_LIST:
	  return "there are no patches";
	  break;
     case PATCH_ID_COLLISION:
	  return "patch with specified id already exists";
	  break;
     case PATCH_ALLOC_FAIL:
	  return "failed to allocate space for patch";
	  break;
     case PATCH_SAMPLE_FAIL:
	  return "failed to load sample file";
	  break;
     case PATCH_NOTE_INVALID:
	  return "specified note is invalid";
	  break;
     case PATCH_FILE_WRITE_FAIL:
	  return "failed to write specified file";
	  break;
     case PATCH_FILE_OPEN_FAIL:
	  return "failed to open specified file";
	  break;
     case PATCH_PAN_INVALID:
	  return "specified panning is invalid";
	  break;
     case PATCH_CHANNEL_INVALID:
	  return "specified channel is invalid";
	  break;
     case PATCH_VOL_INVALID:
	  return "specified volume is invalid";
	  break;
     case PATCH_MASTER_VOLUME_INVALID:
	  return "specified master volume is invalid";
	  break;
     case PATCH_PLAY_MODE_INVALID:
	  return "specified patch play mode is invalid";
	  break;
     case PATCH_LIMIT:
	  return "maximum patch count reached, can't create another";
	  break;
     case PATCH_SAMPLE_INDEX_INVALID:
	  return "specified sample is invalid";
	  break;
     default:
	  return "unknown error";
     }
}

/* makes sure a given patch is active and properly configured */
int patch_verify(int id)
{
     if (id < 0 || id >= PATCH_COUNT)
	  return 0;
     if (!patch[id].active)
	  return 0;

     return 1;
}

/* destroy all patches */
void patch_destroy_all()
{
     int id;

     for (id = 0; id < PATCH_COUNT; id++)
	  patch_destroy(id);

     return;
}
     
/* activates all patches matching given criteria */
void patch_activate(int chan, int note, int vel)
{
     int i, j;
     
     /* process the cuts */
     for (i = 0; i < PATCH_COUNT; i++) {
	  if (patch[i].active && patch[i].channel == chan &&
	      (patch[i].note == note || (patch[i].range > 0 && note >= patch[i].lower_note &&
		   note <= patch[i].upper_note))) {

	       /* it's important that we send stop events before we do
		  start events, otherwise weird incidences where two
		  samples are played with the same cut values end up
		  nullifying each other before they ever get played */
	       if (patch[i].cut > 0) {
		    for (j = 0; j < PATCH_COUNT; j++) {
			 if (patch[j].active && 
			     patch[j].cut_by == patch[i].cut) {

			      patch_ring_feeder->patch = &patch[j];
			      patch_ring_feeder->note = -1; /* this tells patch_mixdown() to shutdown 
							       the indicated patch regardless of its note */
			      patch_ring_feeder->tick = getticks();
			      patch_ring_feeder->event = EV_STOP;
			      patch_ring_feeder = (patch_ring_feeder + 1 >= patch_ring + RINGMAX) ? 
				   patch_ring : patch_ring_feeder + 1;
			 }
		    }
	       }		    
	  }
     }

     /* process play events */
     for (i = 0; i < PATCH_COUNT; i++) {
	  if (patch[i].active && patch[i].channel == chan &&
	      (patch[i].note == note || (patch[i].range > 0 && note >= patch[i].lower_note &&
		   note <= patch[i].upper_note))) {

	       patch_ring_feeder->patch = &patch[i];
	       patch_ring_feeder->note = note;
	       patch_ring_feeder->tick = getticks();
	       patch_ring_feeder->event = EV_PLAY;
	       patch_ring_feeder->velocity = vel / 127.0;
	       patch_ring_feeder = (patch_ring_feeder + 1 >= patch_ring + RINGMAX) ? patch_ring :
		    patch_ring_feeder + 1;
	  }
     }
	      
     return;
}

/* activate a single patch with given id */
void patch_activate_by_id(int id)
{
     if (id < 0 || id >= PATCH_COUNT)
	  return;
     if (!patch[id].active)
	  return;

     patch_ring_feeder->patch = &patch[id];
     patch_ring_feeder->note = patch[id].note;
     patch_ring_feeder->tick = getticks();
     patch_ring_feeder->event = EV_PLAY;
     patch_ring_feeder->velocity = 1.0;
     patch_ring_feeder = (patch_ring_feeder + 1 >= patch_ring + RINGMAX) ? patch_ring :
	  patch_ring_feeder + 1;

     return;
}
     
/* deactivate all active patches matching given criteria */
void patch_deactivate(int chan, int note)
{
     int i;

     for (i = 0; i < PATCH_COUNT; i++) {
	  if (patch[i].active && patch[i].play_mode != PM_NORMAL &&
	      patch[i].channel == chan &&
	      (patch[i].note == note || (patch[i].range > 0 && note >= patch[i].lower_note &&
					 note <= patch[i].upper_note))) {
	       
	       patch_ring_feeder->patch = &patch[i];
	       patch_ring_feeder->note = note;
	       patch_ring_feeder->tick = getticks();
	       patch_ring_feeder->event = EV_STOP;
	       patch_ring_feeder = (patch_ring_feeder + 1 >= patch_ring + RINGMAX) ? patch_ring :
		    patch_ring_feeder + 1;
	  }
     }
     
     return;
}

void patch_deactivate_by_id(int id)
{
     if (id < 0 || id >= PATCH_COUNT)
	  return;
     if (!patch[id].active)
	  return;
     if (patch[id].play_mode == PM_NORMAL)
	  return;

     patch_ring_feeder->patch = &patch[id];
     patch_ring_feeder->note = patch[id].note;
     patch_ring_feeder->tick = getticks();
     patch_ring_feeder->event = EV_STOP;
     patch_ring_feeder = (patch_ring_feeder + 1 >= patch_ring + RINGMAX) ? patch_ring :
	  patch_ring_feeder + 1;
     
     return;
}

/* returns assigned patch id on success, negative value on failure */
int patch_create(const char *name)
{
     int id, i;
     
     /* find an unoccupied patch id */
     for (id = 0; patch[id].active; id++)
	  if (id == PATCH_COUNT)
	       return PATCH_LIMIT;

     patch[id].active = 1;
     
     debug("Creating patch %s, id (%d).\n", name, id);

     strncpy(patch[id].name, name, PATCH_MAX_NAME);
     patch[id].name[PATCH_MAX_NAME-1] = '\0';
     patch[id].sample.name[0] = '\0';
     patch[id].sample.frames = 0;
     patch[id].channel = 0;
     patch[id].note = 60;
     patch[id].range = 0;
     patch[id].lower_note = 60;
     patch[id].upper_note = 60;
     patch[id].volume = 1.0;
     patch[id].pan = 0.5;
     patch[id].play_mode = PM_NORMAL;
     patch[id].cut = 0;
     patch[id].cut_by = 0;
     patch[id].sample_start = 0;
     patch[id].sample_stop = 0;
     patch[id].loop_start = 0;
     patch[id].loop_stop = 0;

     pthread_mutex_lock(&patch[id].sample.mutex);
     patch[id].sample.sp = NULL;
     pthread_mutex_unlock(&patch[id].sample.mutex);

     /* set display_index to next unique value */
     patch[id].display_index = 0;
     for (i = 0; i < PATCH_COUNT; i++) {
	  if (i == id)
	       continue;
	  if (patch[i].active && patch[i].display_index >= patch[id].display_index) {
	       patch[id].display_index = patch[i].display_index + 1;
	  }
     }

     debug("chosen display: %d\n", patch[id].display_index);
     return id;
}

/* destroy a single patch with given id */
int patch_destroy(int id)
{
     int index;

     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     debug("Removing patch: %d\n", id);

     patch[id].active = 0;
     pthread_mutex_lock(&patch[id].sample.mutex);
     if (patch[id].sample.sp != NULL) {
	  free(patch[id].sample.sp);
	  patch[id].sample.sp = NULL;
     }
     pthread_mutex_unlock(&patch[id].sample.mutex);
     
     /* every active patch with a display_index greater than
	this patch's needs to have it's value decremented so
	that we preservere continuity */
     index = patch[id].display_index;
     for (id = 0; id < PATCH_COUNT; id++) {
	  if (patch[id].active && patch[id].display_index > index)
	       patch[id].display_index--;
     }

     return 0;
}

int patch_display_index_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].display_index;
}

int patch_name_set(int id, const char *name)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     strncpy(patch[id].name, name, PATCH_MAX_NAME);
     return 0;
}

char *patch_name_get(int id)
{
     char *name;

     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  name = strdup("\0");
     else
	  name = strdup(patch[id].name);

     return name;
}

char *patch_sample_name_get(int id)
{
     char *name;

     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  name = strdup("\0");
     else
	  name = strdup(patch[id].sample.name);
     return name;
}

/* oh my god, don't fuck up the pointer this returns, for the love of
 * GOD! */
float const *patch_sample_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return NULL;

     return patch[id].sample.sp;
}

int patch_volume_set(int id, int vol)
{

     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (vol < 0 || vol > 100)
	  return PATCH_VOL_INVALID;

     patch[id].volume = vol / 100.0;
     return 0;
}

int patch_volume_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return (int)(patch[id].volume * 100);
}

int patch_channel_set(int id, int channel)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (channel < 0 || channel > 15)
	  return PATCH_CHANNEL_INVALID;

     patch[id].channel = channel;
     return 0;
}

int patch_channel_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].channel;
}

int patch_master_volume_set(int val)
{

     if (val < 0 || val > 100)
	  return PATCH_MASTER_VOLUME_INVALID;

     master_volume = val / 100.0;
     return 0;
}

int patch_note_set(int id, int note)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (note < 0 || note > 127)
	  return PATCH_NOTE_INVALID;

     patch[id].note = note;
     return 0;
}

int patch_note_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].note;
}

int patch_range_set(int id, int range)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     patch[id].range = range;
     return 0;
}

int patch_range_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].range;
}

int patch_lower_note_set(int id, int note)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (note < 0 || note > 127)
	  return PATCH_NOTE_INVALID;

     patch[id].lower_note = note;
     return 0;
}

int patch_lower_note_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].lower_note;
}

int patch_upper_note_set(int id, int note)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (note < 0 || note > 127)
	  return PATCH_NOTE_INVALID;

     patch[id].upper_note = note;
     return 0;
}

int patch_upper_note_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].upper_note;
}

int patch_pan_set(int id, int pan)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (pan < 0 || pan > 256)
	  return PATCH_PAN_INVALID;

     patch[id].pan = pan / 255.0;
     return 0;
}

int patch_pan_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return (int)(patch[id].pan * 255);
}

int patch_play_mode_set(int id, play_mode_t mode)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (mode != PM_NORMAL && 
	 mode != PM_TRIM && 
	 mode != PM_LOOP && 
	 mode != PM_PINGPONG &&
	 mode != PM_REVERSE) {
	 
	  return PATCH_PLAY_MODE_INVALID;
     }

     patch[id].play_mode = mode;
     return 0;
}

play_mode_t patch_play_mode_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].play_mode;
} 

int patch_cut_set(int id, int cut)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     patch[id].cut = cut;
     return 0;
}

int patch_cut_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].cut;
}

int patch_cut_by_set(int id, int cut_by)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     patch[id].cut_by = cut_by;
     return 0;
}

int patch_cut_by_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].cut_by;
}

int patch_sample_start_set(int id, int start)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (patch[id].sample.sp == NULL)
	  return 0;

     if (start < 0)
	  start = 0;

     patch[id].sample_start = start;
     if (start > patch[id].sample_stop)
	  patch[id].sample_stop = start;
     return 0;
}

int patch_sample_start_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].sample_start;
}

int patch_sample_stop_set(int id, int stop)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (patch[id].sample.sp == NULL)
	  return 0;

     if (stop >= patch[id].sample.frames)
	  stop = patch[id].sample.frames-1;

     patch[id].sample_stop = stop;
     if (stop < patch[id].sample_start)
	  patch[id].sample_start = stop;
     return 0;
}

int patch_sample_stop_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].sample_stop;
}

int patch_loop_start_set(int id, int start)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (patch[id].sample.sp == NULL)
	  return 0;

     if (start < patch[id].sample_start)
	  start = patch[id].sample_start;
     else if (start > patch[id].sample_stop)
	  start = patch[id].sample_stop;

     patch[id].loop_start = start;
     if (start > patch[id].loop_stop)
	  patch[id].loop_stop = start;

     return 0;
}

int patch_loop_start_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].loop_start;
}

int patch_loop_stop_set(int id, int stop)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (patch[id].sample.sp == NULL)
	  return 0;

     if (stop > patch[id].sample_stop)
	  stop = patch[id].sample_stop;
     else if (stop < patch[id].sample_start)
	  stop = patch[id].sample_start;

     patch[id].loop_stop = stop;
     if (stop < patch[id].loop_start)
	  patch[id].loop_start = stop;

     return 0;
}

int patch_loop_stop_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     return patch[id].loop_stop;
}

int patch_frames_get(int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;
     if (patch[id].sample.sp == NULL)
	  return 0;

     return patch[id].sample.frames;
}

int patch_count()
{
     int id, count;

     for (id = count = 0; id < PATCH_COUNT; id++)
	  if (patch[id].active) count++;

     return count;
}

/* place all patch ids, sorted in ascending order by display index,
 into array 'id' and return number of patches */
int patch_dump(int **dump)
{
     int i, j, id, count, tmp;

     *dump = NULL;

     /* determine number of patches */
     count = patch_count();

     if (count == 0)
	  return count;

     /* allocate dump */
     *dump = malloc(sizeof(int) * count);
     if (*dump == NULL)
	  return PATCH_ALLOC_FAIL;

     /* place active patches into dump array */
     for (id = i = 0; id < PATCH_COUNT; id++)
	  if (patch[id].active)
	       (*dump)[i++] = id;

     /* sort dump array by display_index in ascending order */
     for (i = 0; i < count; i++) {
	  for (j = i; j < count; j++) {
	       if (patch[(*dump)[j]].display_index < patch[(*dump)[i]].display_index) {
		    tmp = (*dump)[i];
		    (*dump)[i] = (*dump)[j];
		    (*dump)[j] = tmp;
	       }
	  }
     }

     return count;
}

void patch_flush()
{
     int i;

     for (i=0; i < PATCH_POLYPHONY; i++)
	  patch_playlist[i].active = 0;
     
     patch_ring_consumer = patch_ring_feeder;
}

static int sample_load_file(sample_t *sample, char *name)
{
     SNDFILE *sfp;
     SF_INFO sfinfo;
     SRC_DATA data;
     float *tmp;
     double ratio;
     int err, i;
     
     if ((sfp = sf_open(name, SFM_READ, &sfinfo)) == NULL) {
	  debug("libsndfile doesn't like %s\n", name);
	  return PATCH_SAMPLE_FAIL;
     }

     if (sfinfo.channels > 2) {
	  errmsg("Data can't have more than 2 channels\n");
	  sf_close(sfp);
	  return PATCH_SAMPLE_FAIL;
     }

     pthread_mutex_lock(&sample->mutex);

     /* set aside space for samples */
     if ((tmp = malloc(sfinfo.frames * sfinfo.channels * sizeof(float))) == NULL) {
	  errmsg("Unable to allocate space for samples!\n");
	  sf_close(sfp);
	  pthread_mutex_unlock(&sample->mutex);
	  return PATCH_SAMPLE_FAIL;
     }

     /* free existing samples, if any, and point to new space */
     if (sample->sp != NULL)
	  free(sample->sp);
     sample->sp = tmp;
     sample->frames = (int)sfinfo.frames;
     strncpy(sample->name, name, PATCH_MAX_SAMPLE_NAME);
     sample->name[PATCH_MAX_SAMPLE_NAME-1] = '\0';

     /* load sample file into memory */
     if (sf_readf_float(sfp, sample->sp, sfinfo.frames) != sfinfo.frames) {
	  errmsg("Didn't read correct number of frames into memory... this is bad...\n");
     } else {
	  debug("Read %d frames into memory.\n", (int)sfinfo.frames);
     }

     /* resample if necessary */
     if ((ratio = DRIVER->getrate() / (sfinfo.samplerate * 1.0)) != 1.0) {
	  debug("Resampling... ");
	  if ((tmp = malloc(sizeof(float) * sfinfo.frames * sfinfo.channels * ratio)) == NULL) {
	       errmsg("Couldn't malloc() output buffer, unable to samplerate convert\n");
	       sf_close(sfp);
	       pthread_mutex_unlock(&sample->mutex);
	       return PATCH_SAMPLE_FAIL;
	  }
	  data.src_ratio = ratio;
	  data.data_in = sample->sp;
	  data.data_out = tmp;
	  data.input_frames = sfinfo.frames;
	  data.output_frames = sfinfo.frames * ratio;
	  if ((err = src_simple(&data, SRC_SINC_BEST_QUALITY, sfinfo.channels)) != 0) {
	       errmsg("Failed to resample (%s)\n", src_strerror(err));
	       free(tmp);
	       sf_close(sfp);
	       pthread_mutex_unlock(&sample->mutex);
	       return PATCH_SAMPLE_FAIL;
	  }
	  if (sample->sp != NULL)
	       free(sample->sp);
	  sample->sp = tmp;
	  sample->frames = data.output_frames_gen;
	  debug("done.\n");
     }

     /* convert mono samples to stereo if necessary */
     if (sfinfo.channels == 1) {
	  debug("Converting mono to stereo... ");
	  if ((tmp = malloc(sizeof(float) * sample->frames * 2)) == NULL) {
	       errmsg("Couldn't malloc() output buffer, unable to convert to stereo\n");
	       sf_close(sfp);
	       pthread_mutex_unlock(&sample->mutex);
	       return PATCH_SAMPLE_FAIL;
	  }
	  for (i = 0; i < sample->frames; i++) {
	       tmp[2*i] = sample->sp[i];
	       tmp[2*i+1] = sample->sp[i];
	  }
	  if (sample->sp != NULL)
	       free(sample->sp);
	  sample->sp = tmp;
	  debug("done\n");
     }

     sf_close(sfp);
     pthread_mutex_unlock(&sample->mutex);
     return 0;
}

void patch_preview(char *name)
{
     preview.active = 0;
     sample_load_file(&preview.sample, name);
     preview.next_frame = 0;
     preview.active = 1;
}

int patch_sample_load(int id, char *name)
{
     int val;

     debug("Loading sample %s for patch %d\n", name, id);

     if (id < 0 || id >= PATCH_COUNT || !patch[id].active)
	  return PATCH_ID_INVALID;

     val = sample_load_file(&patch[id].sample, name);

     /* set the sample/loop start/stop point appropriately */
     patch[id].sample_start = 0;
     patch[id].sample_stop = patch[id].sample.frames-1;
     patch[id].loop_start = 0;
     patch[id].loop_stop = patch[id].sample.frames-1;

     return val;
}

void patch_sample_unload(int id)
{
     debug("Unloading sample for patch %d\n", id);
     pthread_mutex_lock(&patch[id].sample.mutex);
     
     if (patch[id].sample.sp != NULL) {
	  free(patch[id].sample.sp);
	  patch[id].sample.sp = NULL;
     }

     patch[id].sample.frames = 0;
     patch[id].sample_start = 0;
     patch[id].sample_stop = 0;
     patch[id].loop_start = 0;
     patch[id].loop_stop = 0;
     patch[id].sample.name[0] = '\0';

     pthread_mutex_unlock(&patch[id].sample.mutex);
}

void patch_sample_reload_all()
{
     int id;
     char *name;

     for (id = 0; id < PATCH_COUNT; id++) {
	  if (!patch[id].active)
	       continue;
	  name = patch_sample_name_get(id);
	  patch_sample_load(id, name);
	  if (name != NULL)
	       free(name);
     }

     return;
}
	  
int patch_file_write(char *name)
{
     char tmp[256];
     int val;
     int *id;
     int i, count;
     xmlDocPtr doc;
     xmlNodePtr rootnode;
     xmlNodePtr xmlpatch;

     debug("Writing current bank to file %s\n", name);

     /* create the guts of the doc */
     doc = xmlNewDoc("1.0");
     rootnode = xmlNewDocNode(doc, NULL, (const xmlChar *)"beef", NULL);
     xmlDocSetRootElement(doc, rootnode);

     /* walk the walk */
     count = patch_dump(&id);
     for (i = 0; i < count; i++) {
	  xmlpatch = xmlNewTextChild(rootnode, NULL, (const xmlChar *)"patch", NULL);
	  
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"name", (const xmlChar *)patch[id[i]].name);

	  pthread_mutex_lock(&patch[id[i]].sample.mutex);
	  if (patch[id[i]].sample.sp != NULL)
	       xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"file",
			       (const xmlChar *)patch[id[i]].sample.name);
	  pthread_mutex_unlock(&patch[id[i]].sample.mutex);
	  
	  sprintf(tmp, "%d", patch[id[i]].channel);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"channel", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch[id[i]].note);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"note", (const xmlChar *)tmp);
	  
	  sprintf(tmp, "%d", (int)(patch[id[i]].volume*100));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"volume", (const xmlChar *)tmp);
	  
	  sprintf(tmp, "%d", (int)(patch[id[i]].pan*255));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"pan", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch[id[i]].play_mode);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"play_mode", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch[id[i]].cut);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"cut", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch[id[i]].cut_by);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"cut_by", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch[id[i]].range);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"range", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch[id[i]].lower_note);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"lower_note", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch[id[i]].upper_note);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"upper_note", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch[id[i]].sample_start);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"sample_start", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch[id[i]].sample_stop);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"sample_stop", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch[id[i]].loop_start);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"loop_start", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch[id[i]].loop_stop);
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"loop_stop", (const xmlChar *)tmp);
     }
     
     /* write the file to disk */
     val = xmlSaveFormatFile(name, doc, 1);
     xmlFreeDoc(doc);

     if (val < 0)
	  return PATCH_FILE_WRITE_FAIL;
     return 0;
}

int patch_file_open(char *path)
{
     xmlDocPtr doc;
     xmlNodePtr rootnode;
     xmlNodePtr xmlpatch;
     xmlNodePtr cur;
     xmlChar *key;
     char *name, *file;
     int channel, note, id, volume, pan, cut, cut_by, range, lower_note, upper_note, 
	  sample_start, sample_stop, loop_start, loop_stop;
     play_mode_t mode;

     debug("Loading bank from file %s\n", path);

     doc = xmlParseFile(path);
     if (doc == NULL) {
	  errmsg("Failed to parse %s\n", path);
	  return PATCH_FILE_OPEN_FAIL;
     }

     rootnode = xmlDocGetRootElement(doc);
     if (rootnode == NULL) {
	  errmsg("%s is empty\n", path);
	  xmlFreeDoc(doc);
	  return PATCH_FILE_OPEN_FAIL;
     }

     if (xmlStrcmp(rootnode->name, (const xmlChar *)"beef") != 0) {
	  errmsg("%s is not a valid 'beef' file\n", path);
	  xmlFreeDoc(doc);
	  return PATCH_FILE_OPEN_FAIL;
     }


     patch_destroy_all();
     for (xmlpatch = rootnode->xmlChildrenNode; xmlpatch != NULL;
	  xmlpatch = xmlpatch->next) {

	  name = file = key = NULL;
	  channel = note = id = volume = pan = mode = cut = cut_by = range = lower_note = upper_note = \
	       sample_start = sample_stop = loop_start = loop_stop = -1;
	  if (!xmlStrcmp(xmlpatch->name, (const xmlChar *)"patch")) {
	       for (cur = xmlpatch->xmlChildrenNode; cur != NULL; cur = cur->next) {
		    if (!xmlStrcmp(cur->name, (const xmlChar *)"name")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 name = strdup((const char *)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"file")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 file = strdup((const char *)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"channel")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 channel = atoi((const char *)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"note")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 note = atoi((const char *)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"volume")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 volume = atoi((const char *)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"pan")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 pan = atoi((const char*)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"play_mode")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 mode = atoi((const char*)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"cut")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 cut = atoi((const char*)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"cut_by")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 cut_by = atoi((const char*)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"range")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 range = atoi((const char*)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"lower_note")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 lower_note = atoi((const char*)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"upper_note")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 upper_note = atoi((const char*)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"sample_start")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 sample_start = atoi((const char*)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"sample_stop")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 sample_stop = atoi((const char*)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"loop_start")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 loop_start = atoi((const char*)key);
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"loop_stop")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 loop_stop = atoi((const char*)key);
		    }
		    if (key != NULL) {
			 xmlFree(key);
			 key = NULL;
		    }
	       }
	       if (name != NULL) { /* must have at least a name to be valid */
		    id = patch_create(name);
		    free(name);
		    name = NULL; /* this is paranoid... but so am I */
		    if (file != NULL) {
			 patch_sample_load(id, file);
			 free(file);
			 file = NULL;
		    }
		    if (channel >= 0)
			 patch_channel_set(id, channel);
		    if (note >= 0)
			 patch_note_set(id, note);
		    if (volume >= 0)
			 patch_volume_set(id, volume);
		    if (pan >= 0)
			 patch_pan_set(id, pan);
		    if (mode >= 0)
			 patch_play_mode_set(id, mode);
		    if (cut > 0)
			 patch_cut_set(id, cut);
		    if (cut_by > 0)
			 patch_cut_by_set(id, cut_by);
		    if (range > 0)
			 patch_range_set(id, range);
		    if (lower_note >= 0)
			 patch_lower_note_set(id, lower_note);
		    if (upper_note >= 0)
			 patch_upper_note_set(id, upper_note);
		    if (sample_start >= 0)
			 patch_sample_start_set(id, sample_start);
		    if (sample_stop >= 0)
			 patch_sample_stop_set(id, sample_stop);
		    if (loop_start >= 0)
			 patch_loop_start_set(id, loop_start);
		    if (loop_stop >= 0)
			 patch_loop_stop_set(id, loop_stop);
	       }
	  }
     }
     xmlFreeDoc(doc);
     return 0;
}

/* supposedly this cubically interpolates, or something... I'd settle
   for just a grilled cheese sandwich */
static float cubici(float y0, float y1, float y2, float y3, float x)
{
     return ((y1+y3-(y0+y2))*pow(x,3))+ 
	  ((2*(y0-y1)+y2-y3)*pow(x,2)) + 
	  ((y2-y0)*x) + y1;
}

#define FT 0.0075 /* time in seconds for fade out (Fade Time) */
#define FC 0.1    /* constant factor by which fade reduces volume per second */
#define CHANS 2   /* stereo for now */

void patch_mixdown(float *buf, int frames)
{
     int i, j, k;                         /* standard fare counters */
     int count;                           /* number of frames we will actually write for a given patch */
     int trunc;                           /* used to truncate the integer part of a floating point */
     float sample;                        /* middle man between what we have and what gets written */
     int tc = DRIVER->getrate() * FT;     /* number of frames to fade out over (tail count) */
     float tp = 1.0 / DRIVER->getrate();  /* period in seconds of one frame (tail period) */
     tick_t current_tick;                 /* what was this invocation called at (ass) */
     static tick_t last_tick = 0;         /* current_tick of last invocation */

     current_tick = getticks();

     /* zero out the buffer */
     for (i = 0; i < (frames*CHANS); i++)
	  buf[i] = 0;

     /* check for new events */
     while (patch_ring_consumer != patch_ring_feeder) {
	  if (patch_ring_consumer->tick > last_tick + frames)
	       break;
	  if (patch_ring_consumer->event == EV_PLAY) {
	       for (i = 0; i < PATCH_POLYPHONY; i++) {
		    if (!patch_playlist[i].active)
		    {
			 patch_playlist[i].patch = patch_ring_consumer->patch;
			 patch_playlist[i].note = patch_ring_consumer->note;
			 patch_playlist[i].velocity = patch_ring_consumer->velocity;
			 patch_playlist[i].offset = patch_ring_consumer->tick - last_tick;
			 if (patch_playlist[i].offset < 0)
			      patch_playlist[i].offset = 0;
			 patch_playlist[i].tailo = -1;
			 patch_playlist[i].tailv = 1;
			 patch_playlist[i].active = 1;

			 /* setup pitch scaling */
			 patch_playlist[i].step = pow(2, (patch_ring_consumer->note - patch_playlist[i].patch->note)/12.0);
			 patch_playlist[i].delta = 0.0;

			 if (patch_playlist[i].patch->play_mode == PM_REVERSE) {
			      patch_playlist[i].step *= -1;
			      patch_playlist[i].next_frame = patch_playlist[i].patch->sample_stop-1;
			 } else {
			      patch_playlist[i].next_frame = patch_playlist[i].patch->sample_start;
			 }

			 break;
		    }
	       }
	  } else if (patch_ring_consumer->event == EV_STOP) {
	       for (i = 0; i < PATCH_POLYPHONY; i++) {
		    if (patch_playlist[i].active && 
			patch_playlist[i].patch == patch_ring_consumer->patch &&
			 (patch_playlist[i].note == patch_ring_consumer->note || patch_ring_consumer->note == -1)) {

			 patch_playlist[i].tailo = patch_ring_consumer->tick - last_tick;
			 if (patch_playlist[i].tailo < 0)
			      patch_playlist[i].tailo = 0;
			 patch_playlist[i].tailc = tc; /* arbitrary */
		    }
	       }
	  }
	  patch_ring_consumer = (patch_ring_consumer + 1 >= patch_ring + RINGMAX) ? patch_ring : 
	       patch_ring_consumer + 1;
     }


     /* do the actual stereo mixing */
     for (i = 0; i < PATCH_POLYPHONY; i++) {
	  if (patch_playlist[i].active) {
	       if (pthread_mutex_trylock(&patch_playlist[i].patch->sample.mutex) != 0)
		    continue;

	       if (patch_playlist[i].patch->sample.sp == NULL) {
		    patch_playlist[i].active = 0;
		    pthread_mutex_unlock(&patch_playlist[i].patch->sample.mutex);
		    continue;
	       }


	       /* determine the number of frames to write */
 	       if (patch_playlist[i].patch->play_mode == PM_LOOP ||
		   patch_playlist[i].patch->play_mode == PM_PINGPONG) {
			 
		    count = frames;

	       } else if (patch_playlist[i].patch->play_mode == PM_REVERSE) {
		    if (patch_playlist[i].next_frame - frames + patch_playlist[i].offset <
			patch_playlist[i].patch->sample_start) {
			 
			 count =  patch_playlist[i].next_frame - 
			      patch_playlist[i].patch->sample_start;
		    
		    } else {
			 count = frames;
		    }
	       } else if (frames + patch_playlist[i].next_frame - patch_playlist[i].offset >
			  patch_playlist[i].patch->sample_stop) {
		    
		    count = patch_playlist[i].patch->sample_stop -
			 patch_playlist[i].next_frame;
		    
	       } else {
		    
		    count = frames;
	       }
	       
	       /* make sure that if we are in the process of stopping
		  we don't write any more frames after we've faded
		  out */
	       if (patch_playlist[i].tailo != -1 && 
		   count > patch_playlist[i].tailc + patch_playlist[i].tailo)
	       {
		    count = patch_playlist[i].tailc + patch_playlist[i].tailo;
	       }
		    
#define MVOL    (master_volume)
#define PVOL    (patch_playlist[i].patch->volume)
#define TVOL    (patch_playlist[i].tailv)
#define VVOL    (patch_playlist[i].velocity)
#define TAILO   (patch_playlist[i].tailo)
#define TAILC   (patch_playlist[i].tailc)
#define PAN     (patch_playlist[i].patch->pan)
#define NXTFRM  (patch_playlist[i].next_frame)
#define STEP    (patch_playlist[i].step)
#define ISTEP   ((int)(patch_playlist[i].step))
#define DELTA   (patch_playlist[i].delta)
#define SSTART  (patch_playlist[i].patch->sample_start)
#define SSTOP   (patch_playlist[i].patch->sample_stop)
#define LSTART  (patch_playlist[i].patch->loop_start)
#define LSTOP   (patch_playlist[i].patch->loop_stop)
#define OFFSET  (patch_playlist[i].offset)
#define FRAMES  (patch_playlist[i].patch->sample.frames)
#define MODE    (patch_playlist[i].patch->play_mode)
#define SP      (patch_playlist[i].patch->sample.sp)

	       for (j = OFFSET*CHANS; j < count*CHANS && NXTFRM < FRAMES; j++) {
		    k = j%CHANS; /* this can be tested to determine if
				  * we are currently dealing with the
				  * left sample or the right */


		    if (STEP == 1.0 || STEP == -1.0) {
			 sample = SP[NXTFRM*CHANS+k];
		    } else if (STEP > 0) {
			 sample = cubici((NXTFRM<1)? 0: SP[NXTFRM*CHANS-CHANS+k],
					 
					 SP[NXTFRM*CHANS+k],
					 
					 (NXTFRM+1>SSTOP)? 0: SP[NXTFRM*CHANS+CHANS+k],
					 
					 (NXTFRM+2>SSTOP)? 0: SP[NXTFRM*CHANS+CHANS*2+k],
					 
					 DELTA);
		
		    } else {
			 sample = cubici((NXTFRM>=SSTOP)? 0: SP[NXTFRM*CHANS+CHANS+k],
					 
					 SP[NXTFRM*CHANS+k],
					 
					 (NXTFRM-1<SSTART)? 0: SP[NXTFRM*CHANS-CHANS+k],
					 
					 (NXTFRM-2<SSTART)? 0: SP[NXTFRM*CHANS-CHANS*2+k],
					 
					 abs(DELTA));
		    }
		    
		    sample *= VVOL * PVOL * MVOL * TVOL;

		    /* this part handles panning */
		    if (k == 0) {
			 buf[j]   += sample * ((PAN<=0.5)? 1.0: ((1.0-PAN)/0.5));
			 buf[j+1] += sample * ((PAN<=0.5)? 0.0: ((PAN-0.5)/0.5));
		    } else {
			 buf[j]   += sample * ((PAN>=0.5)? 1.0: (PAN/0.5));
			 buf[j-1] += sample * ((PAN>=0.5)? 0.0: ((0.5-PAN)/0.5));
	
			 /* determine the new value for NXTFRM */
			 DELTA += STEP;
			 if (abs(DELTA) >= 1.0) {
			      trunc = DELTA;
			      NXTFRM += trunc;
			      DELTA -= trunc;
			 }

			 /* determine if it's time to loop */
			 if (MODE == PM_LOOP && NXTFRM > LSTOP) {
			      NXTFRM = LSTART;
			 } else if (MODE == PM_PINGPONG) {
			      if (STEP > 0 && NXTFRM > LSTOP) {
				   STEP *= -1;
				   DELTA *= -1;
				   NXTFRM = LSTOP+STEP;
			      } else if (STEP < 0 && NXTFRM < LSTART) {
				   STEP *= -1;
				   DELTA *= -1;
				   NXTFRM = LSTART+STEP;
			      }
			 }

			 /* determine if it's time to start fading out */
			 if (TAILO == 0) {
			      TVOL = pow(FC, ((tc-TAILC)*tp)/FT*2);
			      TAILC--;
			 } else if (TAILO > 0) {
			      TAILO--;
			 }
		    }
	       }
	       
	       OFFSET -= count;
	       if (OFFSET < 0) OFFSET = 0;
	       if ((MODE == PM_REVERSE && NXTFRM <= SSTART) ||(MODE != PM_REVERSE && NXTFRM >= SSTOP))
		    patch_playlist[i].active = 0;
	       if (TAILO == 0 && TAILC ==0) patch_playlist[i].active = 0;

	       pthread_mutex_unlock(&patch_playlist[i].patch->sample.mutex);
	  }
     }
     
     /* if there is a sample currently being previewed let's throw it into the mix */
     if (preview.active && pthread_mutex_trylock(&preview.sample.mutex) == 0) {
	  if (preview.sample.sp != NULL) {
	       for (i = 0, j = preview.next_frame*CHANS; i < frames*CHANS && j < preview.sample.frames*CHANS; i++, j++)
		    buf[i] += preview.sample.sp[j] * (master_volume);
	       if ((preview.next_frame = j/CHANS) >= preview.sample.frames) {
		    preview.active = 0;
		    if (preview.sample.sp != NULL)
			 free(preview.sample.sp);
		    preview.sample.sp = NULL;
	       }
	  } else {
	       preview.active = 0;
	  }
	  pthread_mutex_unlock(&preview.sample.mutex);
     }

     last_tick = current_tick;
     return;
}
